
from math import ceil

from discord import File, Guild, Member, TextChannel, User
from discord.abc import Messageable
from discord.errors import HTTPException, NotFound
from redbot.core import app_commands, commands
from redbot.core.bot import Red
from redbot.core.commands.converter import parse_relativedelta
from redbot.core.utils.chat_formatting import bold, error, humanize_timedelta, inline

from ..utilities.factory import message_factory
from ..utilities.registry import type_registry
from ..utilities.utils import get_footer_image, log, send_evidenceformat, timedelta_from_relativedelta
from .moderation import Moderation
from .type import Type


def get_icon(bot: Red) -> File:
    cog = bot.get_cog("Aurora")
    if cog:
        return get_footer_image(cog)
    raise ValueError("Aurora cog not found. How was this managed?")

@type_registry.register(key="note")
class Note(Type):
    key="note"
    string="note"
    verb="noted"

@type_registry.register(key="warn")
class Warn(Type):
    key="warn"
    string="warn"
    verb="warned"

@type_registry.register(key="addrole")
class AddRole(Type):
    key="addrole"
    string="addrole"
    verb="added a role to"

@type_registry.register(key="removerole")
class RemoveRole(Type):
    key="removerole"
    string="removerole"
    verb="removed a role from"

@type_registry.register(key="mute")
class Mute(Type):
    key="mute"
    string="mute"
    verb="muted"

@type_registry.register(key="unmute")
class Unmute(Type):
    key="unmute"
    string="unmute"
    verb="unmuted"

@type_registry.register(key="kick")
class Kick(Type):
    key="kick"
    string="kick"
    verb="kicked"

    @classmethod
    async def handler(cls, ctx: commands.Context, target: Member | User, silent: bool, reason: str = None) -> 'Kick':
        """Kick a user."""
        bot = ctx.bot
        response_message = await ctx.send(f"{target.mention} has been {cls.verb}!\n{bold('Reason:')} {inline(reason)}")

        if silent is False:
            try:
                embed = await message_factory(
                    bot=bot,
                    color=await bot.get_embed_color(ctx.channel),
                    guild=ctx.guild,
                    reason=reason,
                    moderation_type=cls(),
                    moderator=ctx.author,
                    duration=None,
                    response=response_message
                )
                await target.send(embed=embed, file=get_icon(bot))
            except HTTPException:
                pass

        await target.kick(reason=f"{str.title(cls.verb)} by {ctx.author.id} for: {reason}")
        moderation = await Moderation.log(
            bot=bot,
            guild_id=ctx.guild.id,
            moderator_id=ctx.author.id,
            moderation_type=cls(),
            target_type='user',
            target_id=target.id,
            role_id=None,
            duration=None,
            reason=reason
        )
        await response_message.edit(content=f"{target.mention} has been {cls.verb}! (Case {inline(f'#{moderation.id}')})\n{bold('Reason:')} {inline(reason)}")
        await log(ctx, moderation.id)
        await send_evidenceformat(ctx, moderation.id)
        return cls
@type_registry.register(key="ban")
class Ban(Type):
    key="ban"
    string="ban"
    verb="banned"

    @classmethod
    async def handler(cls, ctx: commands.Context, target: Member | User, silent: bool, reason: str = None, delete_messages: app_commands.Choice | None = None) -> 'Ban':
        """Ban a user."""
        bot = ctx.bot
        try:
            await ctx.guild.fetch_ban(target)
            await ctx.send(content=error(f"{target.mention} is already {cls.verb}!"), ephemeral=True)
        except NotFound:
            pass

        if delete_messages is None:
            delete_messages_seconds = 0
        else:
            delete_messages_seconds = delete_messages.value

        response_message = await ctx.send(f"{target.mention} has been {cls.verb}!\n{bold('Reason:')} {inline(reason)}")

        if silent is False:
            try:
                embed = await message_factory(
                    bot=bot,
                    color=await bot.get_embed_color(ctx.channel),
                    guild=ctx.guild,
                    reason=reason,
                    moderation_type=cls(),
                    moderator=ctx.author,
                    duration=None,
                    response=response_message
                )
                await target.send(embed=embed, file=get_icon(bot))
            except HTTPException:
                pass

        await ctx.guild.ban(target, reason=f"{str.title(cls.verb)} by {ctx.author.id} for: {reason}", delete_message_seconds=delete_messages_seconds)
        moderation = await Moderation.log(
            bot=bot,
            guild_id=ctx.guild.id,
            moderator_id=ctx.author.id,
            moderation_type=cls(),
            target_type='user',
            target_id=target.id,
            role_id=None,
            duration=None,
            reason=reason
        )
        await response_message.edit(content=f"{target.mention} has been {cls.verb}! (Case {inline(f'#{moderation.id}')})\n{bold('Reason:')} {inline(reason)}")
        await log(ctx, moderation.id)
        await send_evidenceformat(ctx, moderation.id)
        return cls

    @classmethod
    async def resolve_handler(cls, bot: Red, guild: Guild, target: Member | User, reason: str | None = None) -> None:
        try:
            await guild.fetch_ban(user=target)
        except NotFound:
            return
        await guild.unban(user=target, reason=reason)

        try:
            embed = await message_factory(
                bot=bot,
                color=await bot.get_embed_color(guild.channels[0]),
                guild=guild,
                reason=reason,
                moderation_type=Unban(),
                moderator=None,
                duration=None,
                response=None
            )
            await target.send(embed=embed, file=get_icon(bot))
        except HTTPException:
            pass

@type_registry.register(key="tempban")
class Tempban(Ban):
    key="tempban"
    string="tempban"
    verb="tempbanned"

    @classmethod
    async def handler(cls, ctx: commands.Context, target: Member | User, silent: bool, duration: str, reason: str = None, delete_messages: app_commands.Choice | None = None) -> 'Ban':
        """Ban a user."""
        bot = ctx.bot
        try:
            await ctx.guild.fetch_ban(target)
            await ctx.send(content=error(f"{target.mention} is already {Ban.verb}!"), ephemeral=True)
        except NotFound:
            pass

        if delete_messages is None:
            delete_messages_seconds = 0
        else:
            delete_messages_seconds = delete_messages.value

        parsed_time = parse_relativedelta(duration)
        if not parsed_time:
            await ctx.send(content=error("Please provide a valid duration!"), ephemeral=True)
        try:
            parsed_time = timedelta_from_relativedelta(parsed_time)
        except ValueError:
            await ctx.send(content=error("Please provide a valid duration!"), ephemeral=True)

        response_message = await ctx.send(f"{target.mention} has been {cls.verb} for {humanize_timedelta(parsed_time)}!\n{bold('Reason:')} {inline(reason)}")

        if silent is False:
            try:
                embed = await message_factory(
                    bot=bot,
                    color=await bot.get_embed_color(ctx.channel),
                    guild=ctx.guild,
                    reason=reason,
                    moderation_type=cls(),
                    moderator=ctx.author,
                    duration=parsed_time,
                    response=response_message
                )
                await target.send(embed=embed, file=get_icon(bot))
            except HTTPException:
                pass

        await ctx.guild.ban(target, reason=f"{str.title(cls.verb)} by {ctx.author.id} for: {reason} (Duration: {parsed_time})", delete_message_seconds=delete_messages_seconds)
        moderation = await Moderation.log(
            bot=bot,
            guild_id=ctx.guild.id,
            moderator_id=ctx.author.id,
            moderation_type=cls(),
            target_type='user',
            target_id=target.id,
            role_id=None,
            duration=parsed_time,
            reason=reason
        )
        await response_message.edit(content=f"{target.mention} has been {cls.verb} for {humanize_timedelta(parsed_time)}! (Case {inline(f'#{moderation.id}')})\n{bold('Reason:')} {inline(reason)}")
        await log(ctx, moderation.id)
        await send_evidenceformat(ctx, moderation.id)
        return cls

@type_registry.register(key="softban")
class Softban(Type):
    key="softban"
    string="softban"
    verb="softbanned"

    @classmethod
    async def handler(cls, ctx: commands.Context, target: Member | User, silent: bool, reason: str = None, delete_messages: app_commands.Choice | None = None) -> 'Softban':
        """Softban a user."""
        bot = ctx.bot
        try:
            await ctx.guild.fetch_ban(target)
            await ctx.send(content=error(f"{target.mention} is already {Ban.verb}!"), ephemeral=True)
        except NotFound:
            pass

        if delete_messages is None:
            delete_messages_seconds = 0
        else:
            delete_messages_seconds = delete_messages.value

        response_message = await ctx.send(f"{target.mention} has been {cls.verb}!\n{bold('Reason:')} {inline(reason)}")

        if silent is False:
            try:
                embed = await message_factory(
                    bot=bot,
                    color=await bot.get_embed_color(ctx.channel),
                    guild=ctx.guild,
                    reason=reason,
                    moderation_type=cls(),
                    moderator=ctx.author,
                    duration=None,
                    response=response_message
                )
                await target.send(embed=embed, file=get_icon(bot))
            except HTTPException:
                pass

        await ctx.guild.ban(target, reason=f"{str.title(cls.verb)} by {ctx.author.id} for: {reason}", delete_message_seconds=delete_messages_seconds)
        await ctx.guild.unban(target, reason=f"{str.title(cls.verb)} by {ctx.author.id} for: {reason}")
        moderation = await Moderation.log(
            bot=bot,
            guild_id=ctx.guild.id,
            moderator_id=ctx.author.id,
            moderation_type=cls(),
            target_type='user',
            target_id=target.id,
            role_id=None,
            duration=None,
            reason=reason
        )
        await response_message.edit(content=f"{target.mention} has been {cls.verb}! (Case {inline(f'#{moderation.id}')})\n{bold('Reason:')} {inline(reason)}")
        await log(ctx, moderation.id)
        await send_evidenceformat(ctx, moderation.id)
        return cls

@type_registry.register(key="unban")
class Unban(Type):
    key="unban"
    string="unban"
    verb="unbanned"

    @classmethod
    async def handler(cls, ctx: commands.Context, target: Member | User, silent: bool, reason: str = None) -> 'Unban':
        """Unban a user."""
        bot = ctx.bot
        try:
            await ctx.guild.fetch_ban(target)
        except NotFound:
            await ctx.send(content=error(f"{target.mention} is not {Ban.verb}!"), ephemeral=True)
            return

        response_message = await ctx.send(f"{target.mention} has been {cls.verb}!\n{bold('Reason:')} {inline(reason)}")

        if silent is False:
            try:
                embed = await message_factory(
                    bot=bot,
                    color=await bot.get_embed_color(ctx.channel),
                    guild=ctx.guild,
                    reason=reason,
                    moderation_type=cls(),
                    moderator=ctx.author,
                    duration=None,
                    response=response_message
                )
                await target.send(embed=embed, file=get_icon(bot))
            except HTTPException:
                pass

        await ctx.guild.unban(target, reason=f"{str.title(cls.verb)} by {ctx.author.id} for: {reason}")
        moderation = await Moderation.log(
            bot=bot,
            guild_id=ctx.guild.id,
            moderator_id=ctx.author.id,
            moderation_type=cls(),
            target_type='user',
            target_id=target.id,
            role_id=None,
            duration=None,
            reason=reason
        )
        await response_message.edit(content=f"{target.mention} has been {cls.verb}! (Case {inline(f'#{moderation.id}')})\n{bold('Reason:')} {inline(reason)}")
        await log(ctx, moderation.id)
        await send_evidenceformat(ctx, moderation.id)
        return cls

@type_registry.register(key="slowmode")
class Slowmode(Type):
    key="slowmode"
    string="slowmode"
    verb="set the slowmode in"
    channel=True

    @classmethod
    async def handler(cls, ctx: commands.Context, target: Messageable, silent: bool, duration: str, reason: str) -> 'Slowmode': # pylint: disable=unused-argument
        """Set the slowmode in a channel."""
        bot = ctx.bot
        parsed_time = parse_relativedelta(duration)
        if not parsed_time:
            await ctx.send(content=error("Please provide a valid duration!"), ephemeral=True)
        try:
            parsed_time = timedelta_from_relativedelta(parsed_time)
        except ValueError:
            await ctx.send(content=error("Please provide a valid duration!"), ephemeral=True)

        if ceil(parsed_time.total_seconds()) > 21600:
            await ctx.send(content=error("The slowmode duration cannot exceed 6 hours!"), ephemeral=True)
            return

        if isinstance(target, TextChannel):
            await target.edit(slowmode_delay=ceil(parsed_time.total_seconds()))
        moderation = await Moderation.log(
            bot=bot,
            guild_id=ctx.guild.id,
            moderator_id=ctx.author.id,
            moderation_type=cls(),
            target_type='channel',
            target_id=target.id,
            role_id=None,
            duration=parsed_time,
            reason=None
        )
        await ctx.send(f"{ctx.author.mention} has {cls.verb} {target.mention} to {humanize_timedelta(parsed_time)}!\n{bold('Reason:')} {inline(reason)}")
        await log(ctx, moderation.id)
        return cls

@type_registry.register(key="lockdown")
class Lockdown(Type):
    key="lockdown"
    string="lockdown"
    verb="locked down"
    channel=True
